/**
IBM SOFTWARE DISCLAIMER 

Java array package (draft 0.2). Copyright (1998), International Business 
Machines Corporation.

Permission to use, copy, modify and distribute this software for any
noncommercial purpose and without fee is hereby granted, provided that
this copyright and permission notice appear on all copies of the
software. The name of the IBM Corporation may not be used in any
advertising or publicity pertaining to the use of the software. IBM
makes no warranty or representations about the suitability of the
software for any purpose.  It is provided "AS IS" without any express
or implied warranty, including the implied warranties of
merchantability, fitness for a particular purpose and non-infringement.
IBM shall not be liable for any direct, indirect, special or
consequential damages resulting from the loss of use, data or projects,
whether in an action of contract or tort, arising out of or in
connection with the use or performance of this software.
*/

package array;

/**
   doubleArray1D = 1-dimensional array of doubles
   <P>
   The doubleArray1D concrete class implements 1-dimensional arrays
   of doubles. 1-dimensional arrays can either be instantiated directly 
   through constructors, or they can be
   generated by extracting single-dimensional sections from 
   other 1- or higher-dimensional arrays.
*/

public final class doubleArray1D extends doubleArray {

    protected double[] data;
    protected int n0;
    protected int w0, w1;

    /**
       Create a 1-dimensional array of a certain size. 
       Elements are 0-valued.
       @param n		number of elements in array.
       @exception InvalidArrayShapeException extent must be nonnegative
    */
    public doubleArray1D(int n) throws InvalidArrayShapeException {

	if (n < 0) throw new InvalidArrayShapeException();
	data = new double[n];
	if (data == null) throw new InvalidArrayShapeException();
	n0 = n;
	w0 = 1;
	w1 = 0;
	for (int i=0; i<n0; i++) data[i] = 0.0;
    }

    /**
       Create a 1-dimensional array, with size and values defined
       by a Java array of doubles.
       @param a		Java array defining size and values of the array.
       @exception InvalidArrayShapeException Java array must exist
    */
    public doubleArray1D(double[] a) throws InvalidArrayShapeException {

	if (a == null) throw new InvalidArrayShapeException();
	data = new double[a.length];
	if (data == null) throw new InvalidArrayShapeException();
	n0 = a.length;
	w0 = 1;
	w1 = 0;
	for (int i=0; i<n0; i++) data[i] = a[i];
    }

    /**
       Create a 1-dimensional array as a section of another array.
       Internal use within the package only.
       A 1-dimensional array can be empty (have 0 elements).
    */
    protected doubleArray1D(double[] a, Range I) 
	throws ArrayIndexOutOfBoundsException {

	if (I.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (I.max() >= a.length) throw new ArrayIndexOutOfBoundsException();

	data = a;
	n0 = I.length();
	w0 = I.step();
	w1 = I.first();
    }

    /**
       Return the rank of the 1-dimensional array, i.e. 1.
    */
    public int rank() {

	return 1;
    }

    /**
       Return the number of elements in the array.
    */
    public int size() {

	return n0;
    }

    /**
       Return the extent of the array along its i-th axis.
       @param i	array axis (0 <= i < rank())
       @exception InvalidArrayAxisException i must be between 0 and 1
    */
    public int size(int i) throws InvalidArrayAxisException {

	if (i < 0) throw new InvalidArrayAxisException();
	if (i > 0) throw new InvalidArrayAxisException();

	return n0;
    }

    /**
       Return the shape of the array as an int[] I of length 1.
       I[0] = size(0)
    */
    public int[] shape() {

	int[] I = new int[1];
	I[0] = n0;

	return I;
    }

    /**
       Return the range of array indices along its i-th axis.
       @param i array axis (0 <= i < rank())
       @exception InvalidArrayAxisException i must be between 0 and 0
    */
    public Range range(int i) throws InvalidArrayAxisException {

	if (i < 0) throw new InvalidArrayAxisException();
	if (i > 0) throw new InvalidArrayAxisException();

	try {
	    return new Range(0,n0-1,1);
	} catch (InvalidRangeException e) {
	    // This error should not occur
	    return null;
	}
    }

    /**
       Permute the axis of this array.
       For a one-dimensional array, this is the identity operator.
       @param a0 axis to appear as 0th axis (must be 0)
       @exception InvalidArrayAxisException a0 must be 0
    */
    public doubleArray1D permuteAxes(int a0) 
	throws InvalidArrayAxisException {

	if (a0 < 0) throw new InvalidArrayAxisException();
	if (a0 > 0) throw new InvalidArrayAxisException();

	return this;
    }

    /**
       Reshape this one-dimensional array into a two-dimensional array.
       @param m0 extent of resulting array along 0th axis
       @param m1 extent of resulting array along 1st axis
       @exception NonconformingArrayException m0*m1 must be equal to size of this
    */
    public doubleArray2D reshape(int m0, int m1) 
	throws NonconformingArrayException {

	if (m0 < 0) throw new NonconformingArrayException();
	if (m1 < 0) throw new NonconformingArrayException();
	if (m0*m1 != size()) throw new NonconformingArrayException();

	doubleArray2D result = null;
	try {
	    result = new doubleArray2D(m0,m1);

	    for (int i=0; i<m0; i++) {
		for (int j=0; j<m1; j++) {
		    result.setf(i,j,getf(i*m1+j));
		}
	    }
	} catch (InvalidArrayShapeException e) {
	    System.out.println("Internal error in doubleArray1D.reshape(int,int)");
	    System.exit(-1);
	}

	return result;
    }

    /**
       Reshape this one-dimensional array into a three-dimensional array.
       @param m0 extent of resulting array along 0th axis
       @param m1 extent of resulting array along 1st axis
       @param m2 extent of resulting array along 2nd axis
       @exception NonconformingArrayException m0*m1*m2 must be equal to size of this
    */
    public doubleArray3D reshape(int m0, int m1, int m2) 
	throws NonconformingArrayException {

	if (m0 < 0) throw new NonconformingArrayException();
	if (m1 < 0) throw new NonconformingArrayException();
	if (m2 < 0) throw new NonconformingArrayException();
	if (m0*m1*m2 != size()) throw new NonconformingArrayException();

	doubleArray3D result = null;
	try {
	    result = new doubleArray3D(m0,m1,m2);

	    for (int i=0; i<m0; i++) {
		for (int j=0; j<m1; j++) {
		    for (int k=0; k<m2; k++) {
			result.setf(i,j,k,getf(i*m1*m2+j*m2+k));
		    }
		}
	    }
	} catch (InvalidArrayShapeException e) {
	    System.out.println("Internal error in doubleArray1D.reshape(int,int,int)");
	    System.exit(-1);
	}

	return result;
    }

    /**
       Return the value of an element of the array.
       @param index	index[i] = index along i-th axis, 0 <= index[i] < size(i)
       @exception InvalidArrayIndexException index[] must be of length 1
    */
    public double get(int[] index)
	throws InvalidArrayIndexException, ArrayIndexOutOfBoundsException {

	if (index.length != 1) throw new InvalidArrayIndexException();
	int i = index[0];
	return get(i);
    }

    /**
       Set the value of an element of the array.
       @param index	index[i] = index along i-th axis, 0 <= index[i] < size(i)
       @param d		value to which element is set
       @exception InvalidArrayIndexException index[] must be of length 1
    */
    public void set(int [] index, double d) 
	throws InvalidArrayIndexException, ArrayIndexOutOfBoundsException {

	if (index.length != 1) throw new InvalidArrayIndexException();
	int i = index[0];
	set(i,d);
    }

    /**
       Return the value of an element of the array.
       @param i		index along the 0-th axis of the array.
    */
    public double get(int i) throws ArrayIndexOutOfBoundsException {

	if (i < 0) throw new ArrayIndexOutOfBoundsException();
	if (i >= n0) throw new ArrayIndexOutOfBoundsException();

	return data[i*w0+w1];
    }

    /**
       Return the value of an element of the array, no check.
       @param i		index along the 0-th axis of the array.
    */
    protected double getf(int i) {

	return data[i*w0+w1];
    }

    /**
       Return another doubleArray1D with the values of the
       specified elements.
       @param idx	indexes along the 0-th axis of the array.
       @param result	resulting array
       @exception ArrayIndexOutOfBoundsException indices must be valid
       @exception NonconformingArrayException result must have proper shape
    */
    public doubleArray1D get(Index idx,doubleArray1D result)
	throws ArrayIndexOutOfBoundsException,
	       NonconformingArrayException {

	if (idx.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx.max() >= this.n0) throw new ArrayIndexOutOfBoundsException();

	int len = idx.length();
	if (len != result.n0) throw new NonconformingArrayException();

	for (int i=0; i<len; i++) {
	    result.setf(i,getf(idx.elementf(i)));
	}

	return result;
    }

    /**
       Return another doubleArray1D with the values of the
       specified elements.
       @param idx	indexes along the 0-th axis of the array.
       @exception ArrayIndexOutOfBoundsException indices must be valid
    */
    public doubleArray1D get(Index idx) 
	throws ArrayIndexOutOfBoundsException {

	if (idx.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx.max() >= n0) throw new ArrayIndexOutOfBoundsException();

	int len = idx.length();
	doubleArray1D result = null;
	try {
	    result = new doubleArray1D(len);
	    get(idx,result);
	} catch (InvalidArrayShapeException e) {
	    System.out.println("Internal error in doubleArray1D.get(Index)");
	    System.exit(-1);
	} catch (NonconformingArrayException e) {
	    System.out.println("Internal error in doubleArray1D.get(Index)");
	    System.exit(-1);
	} catch (ArrayIndexOutOfBoundsException e) {
	    System.out.println("Internal error in doubleArray1D.get(Index)");
	    System.exit(-1);
	}

	return result;
    }

    /**
       Return another doubleArray1D with the values of the
       specified elements.
       @param idx	indexes along the 0-th axis of the array.
       @param result	resulting array
       @exception ArrayIndexOutOfBoundsException indices must be valid
       @exception NonconformingArrayException result must have proper shape
    */
    public doubleArray1D get(Range idx,doubleArray1D result)
	throws ArrayIndexOutOfBoundsException,
	       NonconformingArrayException {

	if (idx.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx.max() >= this.n0) throw new ArrayIndexOutOfBoundsException();

	int len = idx.length();
	if (len != result.n0) throw new NonconformingArrayException();

	for (int i=0; i<len; i++) {
	    result.setf(i,getf(idx.elementf(i)));
	}

	return result;
    }

    /**
       Return another doubleArray1D with the values of the
       specified elements.
       @param idx	indexes along the 0-th axis of the array.
       @exception ArrayIndexOutOfBoundsException indices must be valid
    */
    public doubleArray1D get(Range idx) 
	throws ArrayIndexOutOfBoundsException {

	if (idx.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx.max() >= n0) throw new ArrayIndexOutOfBoundsException();

	int len = idx.length();
	doubleArray1D result = null;
	try {
	    result = new doubleArray1D(len);
	    get(idx,result);
	} catch (InvalidArrayShapeException e) {
	    System.out.println("Internal error in doubleArray1D.get(Range)");
	    System.exit(-1);
	} catch (NonconformingArrayException e) {
	    System.out.println("Internal error in doubleArray1D.get(Range)");
	    System.exit(-1);
	} catch (ArrayIndexOutOfBoundsException e) {
	    System.out.println("Internal error in doubleArray1D.get(Range)");
	    System.exit(-1);
	}

	return result;
    }

    /**
       Set the value of an element of the array.
       @param i		index along the 0-th axis of the array
       @param val	value to which element is set
    */
    public void set(int i, double val) throws ArrayIndexOutOfBoundsException {

	if (i < 0) throw new ArrayIndexOutOfBoundsException();
	if (i >= n0) throw new ArrayIndexOutOfBoundsException();

	data[i*w0+w1] = val;
    }

    /**
       Set the value of an element of the array, no check
       @param i		index along the 0-th axis of the array
       @param val	value to which element is set
    */
    protected void setf(int i, double val) {

	data[i*w0+w1] = val;
    }

    /**
       Set the value of a list of elements of the array.
       @param idx	indexes along the 0-th axis of the array
       @param val	list of values for elements
       @exception NonconformingArrayException shapes must match
    */
    public void set(Index idx, doubleArray1D val) 
	throws ArrayIndexOutOfBoundsException, 
	       NonconformingArrayException {

	if (idx.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx.max() >= n0) throw new ArrayIndexOutOfBoundsException();

	int len = idx.length();
	if (len != val.n0) throw new NonconformingArrayException();

	for (int i=0; i<len; i++) {
	    setf(idx.elementf(i),val.getf(i));
	}
    }

    /**
       Set the value of a list of elements of the array.
       @param idx	indexes along the 0-th axis of the array
       @param val	value for elements
    */
    public void set(Index idx, double val) 
	throws ArrayIndexOutOfBoundsException {

	if (idx.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx.max() >= n0) throw new ArrayIndexOutOfBoundsException();

	int len = idx.length();

	for (int i=0; i<len; i++) {
	    setf(idx.elementf(i),val);
	}
    }

    /**
       Set the value of a list of elements of the array.
       @param idx	indexes along the 0-th axis of the array
       @param val	list of values for elements
       @exception NonconformingArrayException shapes must match
    */
    public void set(Range idx, doubleArray1D val) 
	throws ArrayIndexOutOfBoundsException, 
	       NonconformingArrayException {

	if (idx.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx.max() >= n0) throw new ArrayIndexOutOfBoundsException();

	int len = idx.length();
	if (len != val.n0) throw new NonconformingArrayException();

	for (int i=0; i<len; i++) {
	    setf(idx.elementf(i),val.getf(i));
	}
    }

    /**
       Set the value of a list of elements of the array.
       @param idx	indexes along the 0-th axis of the array
       @param val	value for elements
    */
    public void set(Range idx, double val) 
	throws ArrayIndexOutOfBoundsException {

	if (idx.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx.max() >= n0) throw new ArrayIndexOutOfBoundsException();

	int len = idx.length();

	for (int i=0; i<len; i++) {
	    setf(idx.elementf(i),val);
	}
    }

    /**
       Extract a 1-dimensional section from the array.
       @param I		range of indices defining the section
       @exception InvalidRangeException	range of indices must be within bounds
    */
    public doubleArray1D section(Range I) throws InvalidRangeException {

	if (I.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (I.max() >= n0) throw new ArrayIndexOutOfBoundsException();

	Range X = new Range(I.first()*w0+w1,I.last()*w0+w1,I.step()*w0);
	return new doubleArray1D(data,X);
    }

    /**
       Extract a 0-dimensional section from the array.
       @param i		index of the element defining the section.
    */
    public doubleArray0D section(int i) {

	if (i < 0) throw new ArrayIndexOutOfBoundsException();
	if (i >= n0) throw new ArrayIndexOutOfBoundsException();

	return new doubleArray0D(data,i*w0+w1);
    }

    /**
       Assign a scalar value to all elements of the array.
       @param d		value to be assigned
    */
    public void assign(double d) {

	try {
	    int n = size(0);
	    for (int i=0; i<n; i++) {
		set(i,d);
	    }
	} catch (InvalidArrayAxisException e) {
	    System.out.println("Internal error in doubleArray1D.assign(double)");
	    System.exit(-1);
	}
    }

    /**
       Assign the values from another one-dimensional array to this array.
       @param a		array with source values
       @exception NonconformingArrayException arrays must be of same shape
    */
    public void assign(doubleArray1D a) 
	throws NonconformingArrayException {

	try {
	    int n = size(0);
	    if (n != a.size(0)) throw new NonconformingArrayException();

	    for (int i=0; i<n; i++) {
		set(i,a.get(i));
	    }
	} catch (InvalidArrayAxisException e) {
	    System.out.println("Internal error in doubleArray1D.assign(doubleArray1D)");
	    System.exit(-1);
	}
    }

    /**
       Replicates an array by adding a dimension.
       @param axis	axis along which replication is performed
       @param copies	number of copies along that axis
       @param result	resulting array
       @exception NonconformingArrayException result array must be of right shape
       @exception InvalidArrayAxisException 0 <= axis <= 1
    */
    public void spread(int axis,int copies,doubleArray2D result)
	throws NonconformingArrayException, InvalidArrayAxisException {

	if (axis < 0) throw new NonconformingArrayException();
	if (axis > 1) throw new NonconformingArrayException();

	if (axis == 0) {
	    /*
	     * Spread this vector as the rows of a 2D array.
	     */
	    int n = size(0);
	    if (n != result.size(1)) throw new NonconformingArrayException();
	    if (copies != result.size(0)) throw new NonconformingArrayException();
	    for (int i=0; i<copies; i++) {
		for (int j=0; j<n; j++) {
		    result.setf(i,j,getf(j));
		}
	    }

	    return;
	} else {
	    /*
	     * Spread this vector as the columns of a 2D array.
	     */
	    int n = size(0);
	    if (n != result.size(0)) throw new NonconformingArrayException();
	    if (copies != result.size(1)) throw new NonconformingArrayException();
	    for (int i=0; i<n; i++) {
		for (int j=0; j<copies; j++) {
		    result.setf(i,j,getf(i));
		}
	    }

	    return;
	}
    }
    /**
       Transforms this doubleArray1D to a double[].
    */
    public double[] toJava() {

	int n = n0;
	double[] result = new double[n];
	for (int i=0; i<n; i++) {
	    result[i] = getf(i);
	}

	return result;
    }
}
