/**
IBM SOFTWARE DISCLAIMER 

Java array package (draft 0.2). Copyright (1998), International Business 
Machines Corporation.

Permission to use, copy, modify and distribute this software for any
noncommercial purpose and without fee is hereby granted, provided that
this copyright and permission notice appear on all copies of the
software. The name of the IBM Corporation may not be used in any
advertising or publicity pertaining to the use of the software. IBM
makes no warranty or representations about the suitability of the
software for any purpose.  It is provided "AS IS" without any express
or implied warranty, including the implied warranties of
merchantability, fitness for a particular purpose and non-infringement.
IBM shall not be liable for any direct, indirect, special or
consequential damages resulting from the loss of use, data or projects,
whether in an action of contract or tort, arising out of or in
connection with the use or performance of this software.
*/

package array;

/**
   doubleArray2D = 2-dimensional array of doubles
   <P>
   The doubleArray2D concrete class implementes 2-dimensional arrays
   of doubles. 2-dimensional arrays can either be instantiated directly 
   through constructors, or they can be
   generated by extracting two-dimensional sections from 
   other 2- or higher-dimensional arrays.
*/

public final class doubleArray2D extends doubleArray {

    protected double[] data;
    protected int n0, n1;
    protected int w0, w1, w2;

    /**
       Create a 2-dimensional array of a certain shape.
       Elements are 0-valued
       @param n		extent of the array along 0th axis
       @param m		extent of the array along 1st axis
       @exception InvalidArrayShapeException all extents must be nonnegative
    */
    public doubleArray2D(int n, int m) throws InvalidArrayShapeException {

	if (n < 0) throw new InvalidArrayShapeException();
	if (m < 0) throw new InvalidArrayShapeException();
	long size = n*m;
	if (size > Integer.MAX_VALUE) throw new OutOfMemoryError();
	data = new double[n*m];
	if (data == null) throw new InvalidArrayShapeException();
	n0 = n; n1 = m;
	w0 = m; w1 = 1; w2 = 0;

    }

    /**
       Create a 2-dimensional array, with shape and values defined by
       a Java double[][] array.
       @param a		Java double[][] array defining shape and values of the array
       @exception InvalidArrayShapeException The Java array must be rectangular
    */
    public doubleArray2D(double[][] a) throws InvalidArrayShapeException {

	if (a == null) throw new InvalidArrayShapeException();
	int n = a.length;
	int m = a[0].length;
	long size = n*m;
	if (size > Integer.MAX_VALUE) throw new OutOfMemoryError();
	data = new double[n*m];
	if (data == null) throw new InvalidArrayShapeException();
	for (int i=0; i<n; i++) {
	    if (a[i].length != m) throw new InvalidArrayShapeException();
	    for (int j=0; j<m; j++) {
		data[i*m+j] = a[i][j];
	    }
	}
	n0 = n; n1 = m;
	w0 = m; w1 = 1; w2 = 0;
    }

    /**
       Create a 2-dimensional array as a section of another array.
       Internal use within the package only.
       A 2-dimensional array can be empty (have 0 elements).
    */
    protected doubleArray2D(double[] a, Range I, Range J) 
	throws ArrayIndexOutOfBoundsException {

	if (I.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (J.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (I.max()+J.max() >= a.length) throw new ArrayIndexOutOfBoundsException();

	data = a;
	n0 = I.length(); n1 = J.length();
	w0 = I.step();   w1 = J.step(); w2 = I.first()+J.first();
    }

    /**
       Return the rank of the 2-dimensional array, i.e. 2.
    */
    public int rank() {

	return 2;
    }

    /**
       Return the number of elements in the array.
    */
    public int size() {

	return n0*n1;
    }

    /**
       Return the extent of the array along its i-th axis.
       @param i		array axis (0 <= i < rank())
       @exception InvalidArrayAxisException i must be between 0 and 1
    */
    public int size(int i) throws InvalidArrayAxisException {

	if (i < 0) throw new InvalidArrayAxisException();
	if (i > 1) throw new InvalidArrayAxisException();

	int shape = 0;
	if (i == 0) shape = n0;
	if (i == 1) shape = n1;
	return shape;
    }

    /**
       Return the shape of the array as an int[] I of length 2.
       I[0] = size(0)
       I[1] = size(1)
    */
    public int[] shape() {

	int[] I = new int[2];
	I[0] = n0;
	I[1] = n1;

	return I;
    }

    /**
       Return the range of array indices along its i-th axis.
       @param i array axis (0 <= i < rank())
       @exception InvalidArrayAxisException i must be between 0 and 1
    */
    public Range range(int i) throws InvalidArrayAxisException {

	if (i < 0) throw new InvalidArrayAxisException();
	if (i > 1) throw new InvalidArrayAxisException();

	Range r = null;
	try {
	    if (i == 0) r = new Range(0,n0-1,1);
	    if (i == 1) r = new Range(0,n1-1,1);
	} catch (InvalidRangeException e) {
	    // This error should not occur
	    return null;
	}

	return r;
    }

    /**
       Permute the axes of this array, thus creating a reference 
       to its transpose.
       @param a0 axis to appear as 0th axis (0 or 1)
       @param a1 axis to appear as 1st axis (0 or 1)
       @exception InvalidArrayAxisException 0 <= a0,a1 <= 1
    */
    public doubleArray2D permuteAxes(int a0,int a1) 
	throws InvalidArrayAxisException {

	if (a0 < 0) throw new InvalidArrayAxisException();
	if (a1 < 0) throw new InvalidArrayAxisException();
	if (a0 > 1) throw new InvalidArrayAxisException();
	if (a1 > 1) throw new InvalidArrayAxisException();

	if ((a0 == 0) && (a1 == 1)) {

	    /*
	     * No permutation, just return this same array.
	     */
	    return this;
	} else if ((a0 == 1) && (a1 == 0)) {

	    /*
	     * Transpose.
	     */
	    Range X = null;
	    Range Y = null;
	    try {
		X = new Range(0+w2,(n1-1)*w1+w2,w1);
		Y = new Range(0,(n0-1)*w0,w0);
	    } catch (InvalidRangeException e) {
		System.out.println("Internal error in doubleArray2D.permuteAxes(int,int)");
		System.exit(-1);
	    }
	    return new doubleArray2D(data,X,Y);
	} else throw new InvalidArrayAxisException();
    }

    /**
       Reshape this two-dimensional array into a one-dimensional array.
       @param m extent of resulting array along 0th axis
       @exception NonconformingArrayException m must be equal to size of this
    */
    public doubleArray1D reshape(int m) 
	throws NonconformingArrayException {

	if (m < 0) throw new NonconformingArrayException();
	if (m != size()) throw new NonconformingArrayException();

	doubleArray1D result = null;
	try {
	    result = new doubleArray1D(m);

	    for (int i=0; i<n0; i++) {
		for (int j=0; j<n1; j++) {
		    result.setf(i*n1+j,getf(i,j));
		}
	    }
	} catch (InvalidArrayShapeException e) {
	    System.out.println("Internal error in doubleArray2D.reshape(int)");
	    System.exit(-1);
	}

	return result;
    }

    /**
       Reshape this two-dimensional array into another two-dimensional array.
       @param m0 extent of resulting array along 0th axis
       @param m1 extent of resulting array along 1st axis
       @exception NonconformingArrayException m0*m1*m2 must be equal to size of this
    */
    public doubleArray2D reshape(int m0, int m1) 
	throws NonconformingArrayException {

	if (m0 < 0) throw new NonconformingArrayException();
	if (m1 < 0) throw new NonconformingArrayException();
	if (m0*m1 != size()) throw new NonconformingArrayException();

	doubleArray2D result = null;
	try {
	    result = reshape(m0*m1).reshape(m0,m1);
	} catch (NonconformingArrayException e) {
	    System.out.println("Internal error in doubleArray2D.reshape(int,int)");
	    System.exit(-1);
	}

	return result;
    }

    /**
       Reshape this two-dimensional array into a three-dimensional array.
       @param m0 extent of resulting array along 0th axis
       @param m1 extent of resulting array along 1st axis
       @param m2 extent of resulting array along 2nd axis
       @exception NonconformingArrayException m0*m1*m2 must be equal to size of this
    */
    public doubleArray3D reshape(int m0, int m1, int m2) 
	throws NonconformingArrayException {

	if (m0 < 0) throw new NonconformingArrayException();
	if (m1 < 0) throw new NonconformingArrayException();
	if (m2 < 0) throw new NonconformingArrayException();
	if (m0*m1*m2 != size()) throw new NonconformingArrayException();

	doubleArray3D result = null;
	try {
	    result = reshape(m0*m1*m2).reshape(m0,m1,m2);
	} catch (NonconformingArrayException e) {
	    System.out.println("Internal error in doubleArray2D.reshape(int,int,int)");
	    System.exit(-1);
	}

	return result;
    }

    /**
       Return the value of an element of the array.
       @param index	index[i] = index along i-th axis, 0 <= index[i] < size(i)
       @exception InvalidArrayIndexException index[] must be of length 2
    */
    public double get(int[] index)
	throws InvalidArrayIndexException, 
	       ArrayIndexOutOfBoundsException {

	if (index.length != 2) throw new InvalidArrayIndexException();
	int i = index[0];
	int j = index[1];
	return get(i,j);
    }

    /**
       Set the value of an element of the array.
       @param index	index[i] = index along i-th axis, 0 <= index[i] < size(i)
       @param d		value to which element is set
       @exception InvalidArrayIndexException index[] must be of length 2
    */
    public void set(int [] index, double d) 
	throws InvalidArrayIndexException, 
	       ArrayIndexOutOfBoundsException {

	if (index.length != 2) throw new InvalidArrayIndexException();
	int i = index[0];
	int j = index[1];
	set(i,j,d);
    }

    /**
       Return the value of an element of the array.
       @param i		index along the 0th axis of the array.
       @param j		index along the 1st axis of the array.
    */
    public double get(int i, int j) 
	throws ArrayIndexOutOfBoundsException {

	if (i < 0) throw new ArrayIndexOutOfBoundsException();
	if (i >= n0) throw new ArrayIndexOutOfBoundsException();
	if (j < 0) throw new ArrayIndexOutOfBoundsException();
	if (j >= n1) throw new ArrayIndexOutOfBoundsException();

	return data[i*w0+j*w1+w2];
    }

    /**
       Return the value of an element of the array, no check.
       @param i		index along the 0th axis of the array.
       @param j		index along the 1st axis of the array.
    */
    protected double getf(int i, int j) 
	throws ArrayIndexOutOfBoundsException {

	return data[i*w0+j*w1+w2];
    }

    /**
       Extract selected elements of this doubleArray2D into a doubleArray1D.
       @param idx0	indexes along the 0th axis of the array
       @param idx1	index along the 1st axis of the array
    */
    public doubleArray1D get(int idx0, Index idx1)
	throws ArrayIndexOutOfBoundsException {

	if (idx0 < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx0 >= n0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.max() >= n1) throw new ArrayIndexOutOfBoundsException();

	int len = idx1.length();
	doubleArray1D result = null;
	try {
	    result = new doubleArray1D(len);
	    get(idx0,idx1,result);
	} catch (InvalidArrayShapeException e) {
	    System.out.println("Internal error in doubleArray2D.get(Index,int)");
	    System.exit(-1);
	} catch (NonconformingArrayException e) {
	    System.out.println("Internal error in doubleArray2D.get(Index,int)");
	    System.exit(-1);
	} catch (ArrayIndexOutOfBoundsException e) {
	    System.out.println("Internal error in doubleArray2D.get(Index,int)");
	    System.exit(-1);
	}

	return result;
    }

    /**
       Extract selected elements of this doubleArray2D into a doubleArray1D.
       @param idx0	indexes along the 0th axis of the array
       @param idx1	index along the 1st axis of the array
       @param result    resulting array
       @exception NonconformingArrayException shapes must match
    */
    public doubleArray1D get(int idx0, Index idx1, doubleArray1D result)
	throws ArrayIndexOutOfBoundsException,
	       NonconformingArrayException {

	if (idx0 < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx0 >= n0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.max() >= n1) throw new ArrayIndexOutOfBoundsException();

	int len = idx1.length();
	try {
	    if (len != result.size(0)) throw new NonconformingArrayException();
	} catch (InvalidArrayAxisException e) {
	    System.out.println("Internal error in doubleArray2D.get(Index,int,doubleArray1D)");
	    System.exit(-1);
	}

	for (int i=0; i<len; i++) {
	    result.setf(i,getf(idx0,idx1.elementf(i)));
	}

	return result;
    }

    /**
       Extract selected elements of this doubleArray2D into a doubleArray1D.
       @param idx0	indexes along the 0th axis of the array
       @param idx1	index along the 1st axis of the array
    */
    public doubleArray1D get(int idx0, Range idx1)
	throws ArrayIndexOutOfBoundsException {

	if (idx0 < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx0 >= n0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.max() >= n1) throw new ArrayIndexOutOfBoundsException();

	int len = idx1.length();
	doubleArray1D result = null;
	try {
	    result = new doubleArray1D(len);
	    get(idx0,idx1,result);
	} catch (InvalidArrayShapeException e) {
	    System.out.println("Internal error in doubleArray2D.get(Index,int)");
	    System.exit(-1);
	} catch (NonconformingArrayException e) {
	    System.out.println("Internal error in doubleArray2D.get(Index,int)");
	    System.exit(-1);
	} catch (ArrayIndexOutOfBoundsException e) {
	    System.out.println("Internal error in doubleArray2D.get(Index,int)");
	    System.exit(-1);
	}

	return result;
    }

    /**
       Extract selected elements of this doubleArray2D into a doubleArray1D.
       @param idx0	indexes along the 0th axis of the array
       @param idx1	index along the 1st axis of the array
       @param result    resulting array
       @exception NonconformingArrayException shapes must match
    */
    public doubleArray1D get(int idx0, Range idx1, doubleArray1D result)
	throws ArrayIndexOutOfBoundsException,
	       NonconformingArrayException {

	if (idx0 < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx0 >= n0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.max() >= n1) throw new ArrayIndexOutOfBoundsException();

	int len = idx1.length();
	try {
	    if (len != result.size(0)) throw new NonconformingArrayException();
	} catch (InvalidArrayAxisException e) {
	    System.out.println("Internal error in doubleArray2D.get(Index,int,doubleArray1D)");
	    System.exit(-1);
	}

	for (int i=0; i<len; i++) {
	    result.setf(i,getf(idx0,idx1.elementf(i)));
	}

	return result;
    }

    /**
       Extract selected elements of this doubleArray2D into a doubleArray1D.
       @param idx0	indexes along the 0th axis of the array
       @param idx1	index along the 1st axis of the array
    */
    public doubleArray1D get(Range idx0, int idx1)
	throws ArrayIndexOutOfBoundsException {

	if (idx0.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx0.max() >= n0) throw new ArrayIndexOutOfBoundsException();
	if (idx1 < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx1 >= n1) throw new ArrayIndexOutOfBoundsException();

	int len = idx0.length();
	doubleArray1D result = null;
	try {
	    result = new doubleArray1D(len);
	    get(idx0,idx1,result);
	} catch (InvalidArrayShapeException e) {
	    System.out.println("Internal error in doubleArray2D.get(Index,int)");
	    System.exit(-1);
	} catch (NonconformingArrayException e) {
	    System.out.println("Internal error in doubleArray2D.get(Index,int)");
	    System.exit(-1);
	} catch (ArrayIndexOutOfBoundsException e) {
	    System.out.println("Internal error in doubleArray2D.get(Index,int)");
	    System.exit(-1);
	}

	return result;
    }

    /**
       Extract selected elements of this doubleArray2D into a doubleArray1D.
       @param idx0	indexes along the 0th axis of the array
       @param idx1	index along the 1st axis of the array
       @param result    resulting array
       @exception NonconformingArrayException shapes must match
    */
    public doubleArray1D get(Range idx0, int idx1, doubleArray1D result)
	throws ArrayIndexOutOfBoundsException,
	       NonconformingArrayException {

	if (idx0.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx0.max() >= n0) throw new ArrayIndexOutOfBoundsException();
	if (idx1 < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx1 >= n1) throw new ArrayIndexOutOfBoundsException();

	int len = idx0.length();
	try {
	    if (len != result.size(0)) throw new NonconformingArrayException();
	} catch (InvalidArrayAxisException e) {
	    System.out.println("Internal error in doubleArray2D.get(Index,int,doubleArray1D)");
	    System.exit(-1);
	}

	for (int i=0; i<len; i++) {
	    result.setf(i,getf(idx0.elementf(i),idx1));
	}

	return result;
    }

    /**
       Extract selected elements of this doubleArray2D into a doubleArray1D.
       @param idx0	indexes along the 0th axis of the array
       @param idx1	index along the 1st axis of the array
    */
    public doubleArray1D get(Index idx0, int idx1)
	throws ArrayIndexOutOfBoundsException {

	if (idx0.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx0.max() >= n0) throw new ArrayIndexOutOfBoundsException();
	if (idx1 < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx1 >= n1) throw new ArrayIndexOutOfBoundsException();

	int len = idx0.length();
	doubleArray1D result = null;
	try {
	    result = new doubleArray1D(len);
	    get(idx0,idx1,result);
	} catch (InvalidArrayShapeException e) {
	    System.out.println("Internal error in doubleArray2D.get(Index,int)");
	    System.exit(-1);
	} catch (NonconformingArrayException e) {
	    System.out.println("Internal error in doubleArray2D.get(Index,int)");
	    System.exit(-1);
	} catch (ArrayIndexOutOfBoundsException e) {
	    System.out.println("Internal error in doubleArray2D.get(Index,int)");
	    System.exit(-1);
	}

	return result;
    }

    /**
       Extract selected elements of this doubleArray2D into a doubleArray1D.
       @param idx0	indexes along the 0th axis of the array
       @param idx1	index along the 1st axis of the array
       @param result    resulting array
       @exception NonconformingArrayException shapes must match
    */
    public doubleArray1D get(Index idx0, int idx1, doubleArray1D result)
	throws ArrayIndexOutOfBoundsException,
	       NonconformingArrayException {

	if (idx0.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx0.max() >= n0) throw new ArrayIndexOutOfBoundsException();
	if (idx1 < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx1 >= n1) throw new ArrayIndexOutOfBoundsException();

	int len = idx0.length();
	try {
	    if (len != result.size(0)) throw new NonconformingArrayException();
	} catch (InvalidArrayAxisException e) {
	    System.out.println("Internal error in doubleArray2D.get(Index,int,doubleArray1D)");
	    System.exit(-1);
	}

	for (int i=0; i<len; i++) {
	    result.setf(i,getf(idx0.elementf(i),idx1));
	}

	return result;
    }

    /**
       Return another doubleArray2D with the values of the 
       specified elements.
       @param idx0	indexes along the 0th axis of the array
       @param idx1	indexes along the 1st axis of the array
    */
    public doubleArray2D get(Index idx0, Index idx1)
	throws ArrayIndexOutOfBoundsException {

	if (idx0.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx0.max() >= n0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.max() >= n1) throw new ArrayIndexOutOfBoundsException();

	int len0 = idx0.length();
	int len1 = idx1.length();
	doubleArray2D result = null;
	try {
	    result = new doubleArray2D(len0,len1);
	    get(idx0,idx1,result);
	} catch (InvalidArrayShapeException e) {
	    System.out.println("Internal error in doubleArray2D.get(Range,Range)");
	    System.exit(-1);
	} catch (NonconformingArrayException e) {
	    System.out.println("Internal error in doubleArray2D.get(Range,Range)");
	    System.exit(-1);
	} catch (ArrayIndexOutOfBoundsException e) {
	    System.out.println("Internal error in doubleArray2D.get(Range,Range)");
	    System.exit(-1);
	}

	return result;
    }

    /**
       Extract selected elements of this doubleArray2D into another
       doubleArray2D.
       @param idx0	indexes along the 0th axis of the array
       @param idx1	indexes along the 1st axis of the array
       @param result	array where elements are stored
       @exception NonconformingArrayException shapes must match
    */
    public doubleArray2D get(Index idx0, Index idx1, doubleArray2D result)
	throws ArrayIndexOutOfBoundsException, 
	       NonconformingArrayException {

	if (idx0.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx0.max() >= n0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.max() >= n1) throw new ArrayIndexOutOfBoundsException();

	int len0 = idx0.length();
	int len1 = idx1.length();

	try {
	    if (len0 != result.size(0)) throw new NonconformingArrayException();
	    if (len1 != result.size(1)) throw new NonconformingArrayException();
	} catch (InvalidArrayAxisException e) {
	    System.out.println("Internal error in doubleArray2D.get(Range,Range,doubleArray2D)");
	    System.exit(-1);
	}

        for (int i=0; i<len0; i++) {
	    int ii = idx0.elementf(i);
	    for (int j=0; j<len1; j++) {
		result.setf(i,j,getf(ii,idx1.elementf(j)));
	    }
	}

	return result;
    }

    /**
       Return another doubleArray2D with the values of the 
       specified elements.
       @param idx0	indexes along the 0th axis of the array
       @param idx1	indexes along the 1st axis of the array
    */
    public doubleArray2D get(Range idx0, Index idx1)
	throws ArrayIndexOutOfBoundsException {

	if (idx0.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx0.max() >= n0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.max() >= n1) throw new ArrayIndexOutOfBoundsException();

	int len0 = idx0.length();
	int len1 = idx1.length();
	doubleArray2D result = null;
	try {
	    result = new doubleArray2D(len0,len1);
	    get(idx0,idx1,result);
	} catch (InvalidArrayShapeException e) {
	    System.out.println("Internal error in doubleArray2D.get(Range,Range)");
	    System.exit(-1);
	} catch (NonconformingArrayException e) {
	    System.out.println("Internal error in doubleArray2D.get(Range,Range)");
	    System.exit(-1);
	} catch (ArrayIndexOutOfBoundsException e) {
	    System.out.println("Internal error in doubleArray2D.get(Range,Range)");
	    System.exit(-1);
	}

	return result;
    }

    /**
       Extract selected elements of this doubleArray2D into another
       doubleArray2D.
       @param idx0	indexes along the 0th axis of the array
       @param idx1	indexes along the 1st axis of the array
       @param result	array where elements are stored
       @exception NonconformingArrayException shapes must match
    */
    public doubleArray2D get(Range idx0, Index idx1, doubleArray2D result)
	throws ArrayIndexOutOfBoundsException, 
	       NonconformingArrayException {

	if (idx0.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx0.max() >= n0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.max() >= n1) throw new ArrayIndexOutOfBoundsException();

	int len0 = idx0.length();
	int len1 = idx1.length();

	try {
	    if (len0 != result.size(0)) throw new NonconformingArrayException();
	    if (len1 != result.size(1)) throw new NonconformingArrayException();
	} catch (InvalidArrayAxisException e) {
	    System.out.println("Internal error in doubleArray2D.get(Range,Range,doubleArray2D)");
	    System.exit(-1);
	}

        for (int i=0; i<len0; i++) {
	    int ii = idx0.elementf(i);
	    for (int j=0; j<len1; j++) {
		result.setf(i,j,getf(ii,idx1.elementf(j)));
	    }
	}

	return result;
    }

    /**
       Return another doubleArray2D with the values of the 
       specified elements.
       @param idx0	indexes along the 0th axis of the array
       @param idx1	indexes along the 1st axis of the array
    */
    public doubleArray2D get(Range idx0, Range idx1)
	throws ArrayIndexOutOfBoundsException {

	if (idx0.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx0.max() >= n0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.max() >= n1) throw new ArrayIndexOutOfBoundsException();

	int len0 = idx0.length();
	int len1 = idx1.length();
	doubleArray2D result = null;
	try {
	    result = new doubleArray2D(len0,len1);
	    get(idx0,idx1,result);
	} catch (InvalidArrayShapeException e) {
	    System.out.println("Internal error in doubleArray2D.get(Range,Range)");
	    System.exit(-1);
	} catch (NonconformingArrayException e) {
	    System.out.println("Internal error in doubleArray2D.get(Range,Range)");
	    System.exit(-1);
	} catch (ArrayIndexOutOfBoundsException e) {
	    System.out.println("Internal error in doubleArray2D.get(Range,Range)");
	    System.exit(-1);
	}

	return result;
    }

    /**
       Extract selected elements of this doubleArray2D into another
       doubleArray2D.
       @param idx0	indexes along the 0th axis of the array
       @param idx1	indexes along the 1st axis of the array
       @param result	array where elements are stored
       @exception NonconformingArrayException shapes must match
    */
    public doubleArray2D get(Range idx0, Range idx1, doubleArray2D result)
	throws ArrayIndexOutOfBoundsException, 
	       NonconformingArrayException {

	if (idx0.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx0.max() >= n0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.max() >= n1) throw new ArrayIndexOutOfBoundsException();

	int len0 = idx0.length();
	int len1 = idx1.length();

	try {
	    if (len0 != result.size(0)) throw new NonconformingArrayException();
	    if (len1 != result.size(1)) throw new NonconformingArrayException();
	} catch (InvalidArrayAxisException e) {
	    System.out.println("Internal error in doubleArray2D.get(Range,Range,doubleArray2D)");
	    System.exit(-1);
	}

        for (int i=0; i<len0; i++) {
	    int ii = idx0.elementf(i);
	    for (int j=0; j<len1; j++) {
		result.setf(i,j,getf(ii,idx1.elementf(j)));
	    }
	}

	return result;
    }

    /**
       Return another doubleArray2D with the values of the 
       specified elements.
       @param idx0	indexes along the 0th axis of the array
       @param idx1	indexes along the 1st axis of the array
    */
    public doubleArray2D get(Index idx0, Range idx1)
	throws ArrayIndexOutOfBoundsException {

	if (idx0.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx0.max() >= n0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.max() >= n1) throw new ArrayIndexOutOfBoundsException();

	int len0 = idx0.length();
	int len1 = idx1.length();
	doubleArray2D result = null;
	try {
	    result = new doubleArray2D(len0,len1);
	    get(idx0,idx1,result);
	} catch (InvalidArrayShapeException e) {
	    System.out.println("Internal error in doubleArray2D.get(Index,Range)");
	    System.exit(-1);
	} catch (NonconformingArrayException e) {
	    System.out.println("Internal error in doubleArray2D.get(Index,Range)");
	    System.exit(-1);
	} catch (ArrayIndexOutOfBoundsException e) {
	    System.out.println("Internal error in doubleArray2D.get(Index,Range)");
	    System.exit(-1);
	}

	return result;
    }

    /**
       Extract selected elements of this doubleArray2D into another
       doubleArray2D.
       @param idx0	indexes along the 0th axis of the array
       @param idx1	indexes along the 1st axis of the array
       @param result	array where elements are stored
       @exception NonconformingArrayException shapes must match
    */
    public doubleArray2D get(Index idx0, Range idx1, doubleArray2D result)
	throws ArrayIndexOutOfBoundsException, 
	       NonconformingArrayException {

	if (idx0.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx0.max() >= n0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.max() >= n1) throw new ArrayIndexOutOfBoundsException();

	int len0 = idx0.length();
	int len1 = idx1.length();

	try {
	    if (len0 != result.size(0)) throw new NonconformingArrayException();
	    if (len1 != result.size(1)) throw new NonconformingArrayException();
	} catch (InvalidArrayAxisException e) {
	    System.out.println("Internal error in doubleArray2D.get(Index,Range,doubleArray2D");
	    System.exit(-1);
	}
        for (int i=0; i<len0; i++) {
	    int ii = idx0.elementf(i);
	    for (int j=0; j<len1; j++) {
		result.setf(i,j,getf(ii,idx1.elementf(j)));
	    }
	}
	return result;
    }
    /**
       Set the value of an element of the array.
       @param i		index along the 0th axis of the array
       @param j		index along the 1st axis of the array
       @param d		value to which element is set
    */
    public void set(int i, int j, double val) 
	throws ArrayIndexOutOfBoundsException {

	if (i < 0) throw new ArrayIndexOutOfBoundsException();
	if (i >= n0) throw new ArrayIndexOutOfBoundsException();
	if (j < 0) throw new ArrayIndexOutOfBoundsException();
	if (j >= n1) throw new ArrayIndexOutOfBoundsException();

	data[i*w0+j*w1+w2] = val;
    }

    /**
       Set the value of an element of the array, no check.
       @param i		index along the 0th axis of the array
       @param j		index along the 1st axis of the array
       @param d		value to which element is set
    */
    protected void setf(int i, int j, double val) 
	throws ArrayIndexOutOfBoundsException {

	data[i*w0+j*w1+w2] = val;
    }

    /**
       Set the value of a list of elements of the array.
       @param idx0	indexes along the 0th axis of the array
       @param idx1	indexes along the 1st axis of the array
       @param val	list of values for elements
       @exception ArrayIndexOutOfBoundsException indices must be valid
       @exception NonconformingArrayException shapes must match
    */
    public void set(int idx0, Index idx1, doubleArray1D val)
	throws ArrayIndexOutOfBoundsException, 
	       NonconformingArrayException {

	if (idx0 < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx0 >= n0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.max() >= n1) throw new ArrayIndexOutOfBoundsException();

	int len = idx1.length();

	try {
	    if (len != val.size(0)) throw new NonconformingArrayException();
	} catch (InvalidArrayAxisException e) {
	    System.out.println("Internal error in doubleArray2D.set(Range,int,doubleArray1D)");
	    System.exit(-1);
	}

	for (int i=0; i<len; i++) {
	    setf(idx0,idx1.elementf(i),val.getf(i));
	}
    }

    /**
       Set the value of a list of elements of the array.
       @param idx0	indexes along the 0th axis of the array
       @param idx1	indexes along the 1st axis of the array
       @param val	list of values for elements
       @exception ArrayIndexOutOfBoundsException indices must be valid
    */
    public void set(int idx0, Index idx1, double val)
 	throws ArrayIndexOutOfBoundsException {

 	if (idx0 < 0) throw new ArrayIndexOutOfBoundsException();
 	if (idx0 >= n0) throw new ArrayIndexOutOfBoundsException();
 	if (idx1.min() < 0) throw new ArrayIndexOutOfBoundsException();
 	if (idx1.max() >= n1) throw new ArrayIndexOutOfBoundsException();

 	int len = idx1.length();

 	for (int i=0; i<len; i++) {
 	    setf(idx0,idx1.elementf(i),val);
 	}
    }

    /**
       Set the value of a list of elements of the array.
       @param idx0	indexes along the 0th axis of the array
       @param idx1	indexes along the 1st axis of the array
       @param val	list of values for elements
       @exception ArrayIndexOutOfBoundsException indices must be valid
       @exception NonconformingArrayException shapes must match
    */
    public void set(int idx0, Range idx1, doubleArray1D val)
	throws ArrayIndexOutOfBoundsException, 
	       NonconformingArrayException {

	if (idx0 < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx0 >= n0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.max() >= n1) throw new ArrayIndexOutOfBoundsException();

	int len = idx1.length();

	try {
	    if (len != val.size(0)) throw new NonconformingArrayException();
	} catch (InvalidArrayAxisException e) {
	    System.out.println("Internal error in doubleArray2D.set(Range,int,doubleArray1D)");
	    System.exit(-1);
	}

	for (int i=0; i<len; i++) {
	    setf(idx0,idx1.elementf(i),val.getf(i));
	}
    }

    /**
       Set the value of a list of elements of the array.
       @param idx0	indexes along the 0th axis of the array
       @param idx1	indexes along the 1st axis of the array
       @param val	list of values for elements
       @exception ArrayIndexOutOfBoundsException indices must be valid
    */
    public void set(int idx0, Range idx1, double val)
 	throws ArrayIndexOutOfBoundsException {

 	if (idx0 < 0) throw new ArrayIndexOutOfBoundsException();
 	if (idx0 >= n0) throw new ArrayIndexOutOfBoundsException();
 	if (idx1.min() < 0) throw new ArrayIndexOutOfBoundsException();
 	if (idx1.max() >= n1) throw new ArrayIndexOutOfBoundsException();

 	int len = idx1.length();

 	for (int i=0; i<len; i++) {
 	    setf(idx0,idx1.elementf(i),val);
 	}
    }

    /**
       Set the value of a list of elements of the array.
       @param idx0	indexes along the 0th axis of the array
       @param idx1	indexes along the 1st axis of the array
       @param val	list of values for elements
       @exception ArrayIndexOutOfBoundsException indices must be valid
       @exception NonconformingArrayException shapes must match
    */
    public void set(Index idx0, int idx1, doubleArray1D val)
	throws ArrayIndexOutOfBoundsException, 
	       NonconformingArrayException {

	if (idx0.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx0.max() >= n0) throw new ArrayIndexOutOfBoundsException();
	if (idx1 < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx1 >= n1) throw new ArrayIndexOutOfBoundsException();

	int len = idx0.length();

	try {
	    if (len != val.size(0)) throw new NonconformingArrayException();
	} catch (InvalidArrayAxisException e) {
	    System.out.println("Internal error in doubleArray2D.set(Index,int,doubleArray1D)");
	    System.exit(-1);
	}

	for (int i=0; i<len; i++) {
	    setf(idx0.elementf(i),idx1,val.getf(i));
	}
    }

    /**
       Set the value of a list of elements of the array.
       @param idx0	indexes along the 0th axis of the array
       @param idx1	indexes along the 1st axis of the array
       @param val	list of values for elements
       @exception ArrayIndexOutOfBoundsException indices must be valid
    */
    public void set(Index idx0, int idx1, double val)
 	throws ArrayIndexOutOfBoundsException {

 	if (idx0.min() < 0) throw new ArrayIndexOutOfBoundsException();
 	if (idx0.max() >= n0) throw new ArrayIndexOutOfBoundsException();
 	if (idx1 < 0) throw new ArrayIndexOutOfBoundsException();
 	if (idx1 >= n1) throw new ArrayIndexOutOfBoundsException();

 	int len = idx0.length();

 	for (int i=0; i<len; i++) {
 	    setf(idx0.elementf(i),idx1,val);
 	}
    }

    /**
       Set the value of a list of elements of the array.
       @param idx0	indexes along the 0th axis of the array
       @param idx1	indexes along the 1st axis of the array
       @param val	list of values for elements
       @exception ArrayIndexOutOfBoundsException indices must be valid
       @exception NonconformingArrayException shapes must match
    */
    public void set(Range idx0, int idx1, doubleArray1D val)
	throws ArrayIndexOutOfBoundsException, 
	       NonconformingArrayException {

	if (idx0.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx0.max() >= n0) throw new ArrayIndexOutOfBoundsException();
	if (idx1 < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx1 >= n1) throw new ArrayIndexOutOfBoundsException();

	int len = idx0.length();

	try {
	    if (len != val.size(0)) throw new NonconformingArrayException();
	} catch (InvalidArrayAxisException e) {
	    System.out.println("Internal error in doubleArray2D.set(Range,int,doubleArray1D)");
	    System.exit(-1);
	}

	for (int i=0; i<len; i++) {
	    setf(idx0.elementf(i),idx1,val.getf(i));
	}
    }

    /**
        Set the value of a list of elements of the array.
        @param idx0	indexes along the 0th axis of the array
        @param idx1	indexes along the 1st axis of the array
        @param val	list of values for elements
    */
    public void set(Range idx0, int idx1, double val)
 	throws ArrayIndexOutOfBoundsException {

 	if (idx0.min() < 0) throw new ArrayIndexOutOfBoundsException();
 	if (idx0.max() >= n0) throw new ArrayIndexOutOfBoundsException();
 	if (idx1 < 0) throw new ArrayIndexOutOfBoundsException();
 	if (idx1 >= n1) throw new ArrayIndexOutOfBoundsException();

 	int len = idx0.length();

 	for (int i=0; i<len; i++) {
 	    setf(idx0.elementf(i),idx1,val);
 	}
    }

    /**
       Set the value of a list of elements of the array.
       @param idx0	indexes along the 0th axis of the array
       @param idx1	indexes along the 1st axis of the array
       @param val	value for elements
    */
    public void set(Index idx0, Index idx1, double val)
	throws ArrayIndexOutOfBoundsException {

	if (idx0.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx0.max() >= n0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.max() >= n1) throw new ArrayIndexOutOfBoundsException();

	int len0 = idx0.length();
	int len1 = idx1.length();
	for (int i=0; i<len0; i++) {
	    int ii = idx0.elementf(i);
	    for (int j=0; j<len1; j++) {
		setf(ii,idx1.elementf(j),val);
	    }
	}

	return;
    }

    /**
       Set the value of a list of elements of the array.
       @param idx0	indices along the 0th axis of the array
       @param idx1	indices along the 1st axis of the array
       @param val	list of values for elements
       @exception ArrayIndexOutOfBoundsException indices must be valid
       @exception NonconformingArrayException shapes must match
    */
    public void set(Index idx0, Index idx1, doubleArray2D val)
	throws ArrayIndexOutOfBoundsException,
	       NonconformingArrayException {

	if (idx0.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx0.max() >= n0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.max() >= n1) throw new ArrayIndexOutOfBoundsException();

	int len0 = idx0.length();
	int len1 = idx1.length();

	try {
	    if (len0 != val.size(0)) throw new NonconformingArrayException();
	    if (len1 != val.size(1)) throw new NonconformingArrayException();
	} catch (InvalidArrayAxisException e) {
	    System.out.println("Internal error in doubleArray2D.set(Range,Range,doubleArray2D)");
	    System.exit(-1);
	}

	for (int i=0; i<len0; i++) {
	    int ii = idx0.elementf(i);
	    for (int j=0; j<len1; j++) {
		setf(ii,idx1.elementf(j),val.getf(i,j));
	    }
	}

	return;
    }

    /**
       Set the value of a list of elements of the array.
       @param idx0	indices along the 0th axis of the array
       @param idx1	indices along the 1st axis of the array
       @param val	list of values for elements
       @exception ArrayIndexOutOfBoundsException indices must be valid
       @exception NonconformingArrayException shapes must match
    */
    public void set(Range idx0, Index idx1, doubleArray2D val)
	throws ArrayIndexOutOfBoundsException,
	       NonconformingArrayException {

	if (idx0.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx0.max() >= n0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.max() >= n1) throw new ArrayIndexOutOfBoundsException();

	int len0 = idx0.length();
	int len1 = idx1.length();

	try {
	    if (len0 != val.size(0)) throw new NonconformingArrayException();
	    if (len1 != val.size(1)) throw new NonconformingArrayException();
	} catch (InvalidArrayAxisException e) {
	    System.out.println("Internal error in doubleArray2D.set(Range,Range,doubleArray2D)");
	    System.exit(-1);
	}

	for (int i=0; i<len0; i++) {
	    int ii = idx0.elementf(i);
	    for (int j=0; j<len1; j++) {
		setf(ii,idx1.elementf(j),val.getf(i,j));
	    }
	}

	return;
    }

    /**
       Set the value of a list of elements of the array.
       @param idx0	indexes along the 0th axis of the array
       @param idx1	indexes along the 1st axis of the array
       @param val	value for elements
       @exception ArrayIndexOutOfBoundsException indices must be valid
    */
    public void set(Range idx0, Index idx1, double val)
	throws ArrayIndexOutOfBoundsException {

	if (idx0.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx0.max() >= n0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.max() >= n1) throw new ArrayIndexOutOfBoundsException();

	int len0 = idx0.length();
	int len1 = idx1.length();
	for (int i=0; i<len0; i++) {
	    int ii = idx0.elementf(i);
	    for (int j=0; j<len1; j++) {
		setf(ii,idx1.elementf(j),val);
	    }
	}

	return;
    }

    /**
       Set the value of a list of elements of the array.
       @param idx0	indices along the 0th axis of the array
       @param idx1	indices along the 1st axis of the array
       @param val	list of values for elements
       @exception ArrayIndexOutOfBoundsException indices must be valid
       @exception NonconformingArrayException shapes must match
    */
    public void set(Range idx0, Range idx1, doubleArray2D val)
	throws ArrayIndexOutOfBoundsException,
	       NonconformingArrayException {

	if (idx0.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx0.max() >= n0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.max() >= n1) throw new ArrayIndexOutOfBoundsException();

	int len0 = idx0.length();
	int len1 = idx1.length();

	try {
	    if (len0 != val.size(0)) throw new NonconformingArrayException();
	    if (len1 != val.size(1)) throw new NonconformingArrayException();
	} catch (InvalidArrayAxisException e) {
	    System.out.println("Internal error in doubleArray2D.set(Range,Range,doubleArray2D)");
	    System.exit(-1);
	}

	for (int i=0; i<len0; i++) {
	    int ii = idx0.elementf(i);
	    for (int j=0; j<len1; j++) {
		setf(ii,idx1.elementf(j),val.getf(i,j));
	    }
	}

	return;
    }

    /**
       Set the value of a list of elements of the array.
       @param idx0	indexes along the 0th axis of the array
       @param idx1	indexes along the 1st axis of the array
       @param val	value for elements
       @exception ArrayIndexOutOfBoundsException indices must be valid
    */
    public void set(Range idx0, Range idx1, double val)
	throws ArrayIndexOutOfBoundsException {

	if (idx0.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx0.max() >= n0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.max() >= n1) throw new ArrayIndexOutOfBoundsException();

	int len0 = idx0.length();
	int len1 = idx1.length();
	for (int i=0; i<len0; i++) {
	    int ii = idx0.elementf(i);
	    for (int j=0; j<len1; j++) {
		setf(ii,idx1.elementf(j),val);
	    }
	}

	return;
    }

    /**
       Set the value of a list of elements of the array.
       @param idx0	indices along the 0th axis of the array
       @param idx1	indices along the 1st axis of the array
       @param val	list of values for elements
       @exception ArrayIndexOutOfBoundsException indices must be valid
       @exception NonconformingArrayException shapes must match
    */
    public void set(Index idx0, Range idx1, doubleArray2D val)
	throws ArrayIndexOutOfBoundsException,
	       NonconformingArrayException {

	if (idx0.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx0.max() >= n0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.max() >= n1) throw new ArrayIndexOutOfBoundsException();

	int len0 = idx0.length();
	int len1 = idx1.length();

	try {
	    if (len0 != val.size(0)) throw new NonconformingArrayException();
	    if (len1 != val.size(1)) throw new NonconformingArrayException();
	} catch (InvalidArrayAxisException e) {
	    System.out.println("Internal error in doubleArray2D.set(Index,Range,doubleArray2D)");
	    System.exit(-1);
	}

	for (int i=0; i<len0; i++) {
	    int ii = idx0.elementf(i);
	    for (int j=0; j<len1; j++) {
		setf(ii,idx1.elementf(j),val.getf(i,j));
	    }
	}

	return;
    }

    /**
       Set the value of a list of elements of the array.
       @param idx0	indexes along the 0th axis of the array
       @param idx1	indexes along the 1st axis of the array
       @param val	value for elements
    */
    public void set(Index idx0, Range idx1, double val)
	throws ArrayIndexOutOfBoundsException {

	if (idx0.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx0.max() >= n0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (idx1.max() >= n1) throw new ArrayIndexOutOfBoundsException();
	int len0 = idx0.length();
	int len1 = idx1.length();
	for (int i=0; i<len0; i++) {
	    int ii = idx0.elementf(i);
	    for (int j=0; j<len1; j++) {
		setf(ii,idx1.elementf(j),val);
	    }
	}
	return;
    }
    /**
       Extract a 2-dimensional section from the array.
       @param I		range of indices along 0th axis defining the section
       @param J		range of indices along 1st axis defining the section
       @exception InvalidRangeException	ranges of indices must be within bounds of axes
    */
    public doubleArray2D section(Range I, Range J) throws InvalidRangeException {

	if (I.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (I.max() >= n0) throw new ArrayIndexOutOfBoundsException();
	if (J.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (J.max() >= n1) throw new ArrayIndexOutOfBoundsException();

	Range X = new Range(I.first()*w0+w2,I.last()*w0+w2,I.step()*w0);
	Range Y = new Range(J.first()*w1,J.last()*w1,J.step()*w1);
	return new doubleArray2D(data,X,Y);
    }

    /**
       Extract a 1-dimensional section from the array.
       @param I		range of indices along 0th axis defining the section
       @param j		index along 1st axis defining the section
       @exception InvalidRangeException	range of indices must be within bounds of 0th axis
    */
    public doubleArray1D section(Range I, int j) throws InvalidRangeException {

	if (I.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (I.max() >= n0) throw new ArrayIndexOutOfBoundsException();
	if (j < 0) throw new ArrayIndexOutOfBoundsException();
	if (j >= n1) throw new ArrayIndexOutOfBoundsException();

	Range Y = new Range(I.first()*w0+j*w1+w2,I.last()*w0+j*w1+w2,I.step()*w0);
	return new doubleArray1D(data,Y);
    }

    /**
       Extract a 1-dimensional section from the array.
       @param i		index along 0th axis defining the section
       @param J		range of indices along 1st axis defining the section
       @exception InvalidRangeException	range of indices must be within bounds of 1st axis
    */
    public doubleArray1D section(int i, Range J) throws InvalidRangeException {

	if (i < 0) throw new ArrayIndexOutOfBoundsException();
	if (i >= n0) throw new ArrayIndexOutOfBoundsException();
	if (J.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (J.max() >= n1) throw new ArrayIndexOutOfBoundsException();

	Range Y = new Range(i*w0+J.first()*w1+w2,i*w0+J.last()*w1+w2,J.step()*w1);
 	return new doubleArray1D(data,Y);
    }

    /**
       Extract a 0-dimensional section from the array.
       @param i		index along 0th axis defining the section.
       @param j		index along 1st axis defining the section.
    */
    public doubleArray0D section(int i, int j) {

	if (i < 0) throw new ArrayIndexOutOfBoundsException();
	if (i >= n0) throw new ArrayIndexOutOfBoundsException();
	if (j < 0) throw new ArrayIndexOutOfBoundsException();
	if (j >= n1) throw new ArrayIndexOutOfBoundsException();

	return new doubleArray0D(data,i*w0+j*w1+w2);
    }

    /**
       Assign a scalar value to all elements of the array.
       @param d		value to be assigned
    */
    public void assign(double d) {

	try {
	    int n = size(0);
	    int m = size(1);
	    for (int i=0; i<n; i++) {
		for (int j=0; j<m; j++) {
		    setf(i,j,d);
		}
	    }
	} catch (InvalidArrayAxisException e) {
	    System.out.println("Unexpected internal error");
	    System.exit(-1);
	}
    }
    /**
       Assign the values from another two-dimensional array to this array.
       @param a		array with source values
       @exception NonconformingArrayException arrays must be of same shape
    */
    public void assign(doubleArray2D a) 
	throws NonconformingArrayException {

	try {
	    int n = size(0);
	    int m = size(1);

	    if (n != a.size(0)) throw new NonconformingArrayException();
	    if (m != a.size(1)) throw new NonconformingArrayException();

	    for (int i=0; i<n; i++) {
		for (int j=0; j<m; j++) {
		    setf(i,j,a.getf(i,j));
		}
	    }
	} catch (InvalidArrayAxisException e) {
	    System.out.println("Unexpected internal error");
	    System.exit(-1);
	}
    }

    /**
       Divide elements of this array by the corresponding elements 
       of another array b.
       @param a		array with divisor values
       @exception NonconformingArrayException b and this must have same shape
    */
    public doubleArray2D divide(doubleArray2D b) 
       throws NonconformingArrayException {

	doubleArray2D a = this;
	doubleArray2D c = null;
	try {
	    int m = a.size(0);
	    int n = a.size(1);

	    if (n != b.size(0)) throw new NonconformingArrayException();
	    if (m != b.size(1)) throw new NonconformingArrayException();

	    c = new doubleArray2D(m,n);

	    for (int i=0; i<m; i++) {
		for (int j=0; j<n; j++) {
		  c.setf(i,j,a.getf(i,j)/b.getf(i,j));
		}
	    }
	} catch (InvalidArrayAxisException e) {
	    System.out.println("Unexpected internal error");
	    System.exit(-1);
	} catch (InvalidArrayShapeException e) {
	    System.out.println("Unexpected internal error");
	    System.exit(-1);
	}

	return c;
    }
    /**
       Transforms this doubleArray2D to a double[].
    */
    public double[][] toJava() {

	int m = n0;
	int n = n1;
	double[][] result = new double[m][n];
	for (int i=0; i<m; i++) {
	    for (int j=0; j<n; j++) {
		result[i][j] = getf(i,j);
	    }
	}

	return result;
    }

}
